/*************************************************************
 * File 	: kernel/ktau/ktau_hash.c
 * Version 	: $Id: ktau_hash.c,v 1.2 2005/08/29 20:16:00 suravee Exp $
 *
 ************************************************************/

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/linkage.h>      /* system call stuff */
#include <linux/sched.h>        /* task_struct def. */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/interrupt.h>
#include <linux/spinlock.h>     /* spinlock */
#include <linux/rwsem.h>        /* R/W semaphore */
//#include <linux/hardirq.h>	/* This might be needed for 2.6 */

#include <asm/current.h>	/* get_current */
#include <asm/uaccess.h>

#define TAU_NAME "ktau_hash"
#define PFX TAU_NAME

//#define TAU_DEBUG		/* to turn on debug messages */
#include <linux/ktau/ktau_print.h>
#include <linux/ktau/ktau_hash.h>
#include <linux/ktau/ktau_api.h>
#include <linux/ktau/ktau_proc.h>

#define for_each_process(p) for_each_task(p)
extern rwlock_t tasklist_lock;

/*----------------------- KTAU GLOBAL SEMAPHORE -----------------------*/
struct rw_semaphore ktau_global_sem;
void ktau_global_semaphore_init(void){
	init_rwsem(&ktau_global_sem);
}

/*--------------------------- IN API ---------------------------------*/
/*
 * This function initialize ktau_prof structure in task_struct.
 * 	1. Allocate the hash-table which contains pointer 
 * 	   to profile information (h_ent).
 * 	2. Initialize spinlock.
 * 	3. Create the corresponding entry in /proc/<pid>/ktau.
 */
void create_task_profile(struct task_struct *tsk){

	if(!(tsk->ktau = kmalloc(sizeof(ktau_prof),GFP_KERNEL))){
                err("Can't allocate memory for ktau_prof.");
	}else{
		memset(tsk->ktau->ktau_hash,0,KTAU_HASH_SIZE * sizeof(h_ent*));
	}
	/* Initialize entry counter */
	tsk->ktau->ent_count 	= 0;
	tsk->ktau->prev_start 	= 0;
	tsk->ktau->prev_stop	= 0;
	tsk->ktau->last_level	= 0;
	tsk->ktau->cur_level	= 0;
	memset(&tsk->ktau->accum,0,
			sizeof(unsigned long long)*KTAU_NESTED_LEVEL);
	tsk->ktau->ppstate	= NULL;

#ifdef CONFIG_KTAU_TRACE
	tsk->ktau->trbuf	= NULL;
	tsk->ktau->cur_trace	= NULL;
	tsk->ktau->head_trace	= NULL;
	trace_buf_init(tsk);	
#endif /*CONFIG_KTAU_TRACE*/

#ifdef CONFIG_KTAU_PERF_COUNTER
	memset(&tsk->ktau->accum_pcounter,0,
		sizeof(unsigned int)*KTAU_MAX_PERF_COUNTER*KTAU_NESTED_LEVEL);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

	/* Initialize hash table semaphore */
	spin_lock_init(&tsk->ktau->hash_lock);

}

/*
 * This function handles clean up for ktau_prof structure
 * in task_struct.
 * 	1. Freeing all the allocated profiles.
 */
void remove_task_profile(struct task_struct *tsk){
        int i;
	unsigned long kflags = 0;

	ktau_spin_lock(&tsk->ktau->hash_lock,kflags);

	/* Free the memory allocated for each entry */
	for(i=0;i<KTAU_HASH_SIZE;i++){
		if((tsk->ktau->ktau_hash[i])){
			kfree((tsk->ktau->ktau_hash[i]));
			tsk->ktau->ktau_hash[i] = NULL;	
		}
        }
	tsk->ktau->ppstate = NULL;

	ktau_spin_unlock(&tsk->ktau->hash_lock,kflags);

#ifdef CONFIG_KTAU_TRACE
	/* Free ktau_prof */
	trace_buf_destroy(tsk);
#endif /*CONFIG_KTAU_TRACE*/
	kfree(tsk->ktau);
	tsk->ktau = NULL;
}

/*--------------------------- OUT API ---------------------------------*/
/*
 * This function handles dumping the profile hash-table
 */
int dump_hash_self(ktau_prof *buf, 
		ktau_output *out, 
		unsigned long long tol)
{
        int i;
	unsigned long kflags = 0;
	h_ent* h_ptr; 
	o_ent* o_ptr; 

	if(out == NULL)
		return -1;

	out->size = 0;
	out->ent_lst = kmalloc(sizeof(o_ent) * buf->ent_count , GFP_KERNEL);

	ktau_spin_lock(&buf->hash_lock,kflags);

	/* Searching through hash table */
	for(i=0;i<KTAU_HASH_SIZE;i++){
		h_ptr = (h_ent*)(buf->ktau_hash[i]);
		if(h_ptr){
			/*			
			printk(KERN_INFO "dump_hash_self: Entry %4d: addr %x, count %4u, start %4llu, incl %4llu, excl %4llu\n",
				i,
				h_ptr->addr, 
				h_ptr->data.timer.count,
				h_ptr->data.timer.start,
				h_ptr->data.timer.incl,
				h_ptr->data.timer.excl);
			*/	
			/* Copy data over */
			o_ptr = (out->ent_lst + out->size);
			o_ptr->index = i;
			memcpy(&(o_ptr->entry),h_ptr, sizeof(h_ent));
			out->size++;
			/* Check if all data are coppied over. */
			if(out->size >= buf->ent_count) 
				break;
		}
        }

	ktau_spin_unlock(&buf->hash_lock,kflags);
	return 0;
}

/* 
 * This function returns number of entries for 
 * "self".
 */ 
int get_profile_size_self(ktau_prof *buf)
{
	int count	= 0;
	int i		= 0;

	for(i=0;i<KTAU_HASH_SIZE;i++){
		if(buf->ktau_hash[i])
			count++;
	}
	
	return(count);
}

/* 
 * This function returns number of entries for 
 * "many".
 */ 
int get_profile_size_many(pid_t *pid_lst, int* num_pid)
{
	struct task_struct *cur_tsk;
	int count	= 0;
	int i		= 0;
	int cur_count   = 0;
	
	/* Case for all */
	if(*num_pid == 0){
		read_lock(&tasklist_lock);
		for_each_process(cur_tsk){
			*num_pid = (*num_pid) + 1;
			for(i=0;i<KTAU_HASH_SIZE;i++){
				if(cur_tsk->ktau->ktau_hash[i])
					count++;
			}
		}
		read_unlock(&tasklist_lock);	
	}
	/* Case for many pids */	
	else{
		read_lock(&tasklist_lock);
#if 0
		for_each_process(cur_tsk){
			if(cur_tsk->pid == pid_lst[cur_count]){
				/* The specified PID is found */
				for(i=0;i<KTAU_HASH_SIZE;i++){
					if(cur_tsk->ktau->ktau_hash[i])
						count++;
				}
				cur_count++;
			}else if(cur_tsk->pid > pid_lst[cur_count]){
				/* The specified PID is not found */
				cur_count++;
			}
		}
#endif
		for_each_process(cur_tsk){
			do{
				if(cur_tsk->pid == pid_lst[cur_count]){
					/* The specified PID is found */
					
					for(i=0;i<KTAU_HASH_SIZE;i++){
						if(cur_tsk->ktau->ktau_hash[i])
							count++;
					}

					/* Move to the next pid in the list */
					cur_count++;
				
				}else if(cur_tsk->pid > pid_lst[cur_count]){
					/* The specified PID is not found */
					/* Move to the next pid in the list */
					err("get_profile_size_many : WARNING! PID %d not found.\n",
							pid_lst[cur_count]);	
					cur_count++;
				}
				/* Must check for array out-of-bound 
				 * before continue */
				if(cur_count >=  *num_pid) 
					goto done;

			}while(cur_tsk->pid >= pid_lst[cur_count]);
		}
done:
		read_unlock(&tasklist_lock);	
	}
	return(count);
}
/*
 * This function handles purging the profile hash-table
 */
int purge_hash_self(ktau_prof *buf, 
		unsigned long long tol)
{
#if 0
        int i;
	unsigned long flags = 0;
	h_ent* h_ptr;

	ktau_spin_lock(&buf->hash_lock,flags);
		for(i=0;i<KTAU_HASH_SIZE;i++){
			h_ptr = (h_ent*)(buf->ktau_hash[i]);
			if(h_ptr){
				h_ptr->data.timer.count = 0;	
				h_ptr->data.timer.start = 0;	
				h_ptr->data.timer.incl = 0;
				h_ptr->data.timer.excl = 0;
			}
		}
	ktau_spin_unlock(&buf->hash_lock,flags);
#endif
	return 0;
}
/*-------------------------------------------------------------------*/
#if 0 
/* Replaced by dump_hash_many and purge_hash_many */
/*
 * Wrapper of dump_hash_self w/ locking 
 */
int dump_hash_other(ktau_prof *buf, 
		ktau_output *out, 
		unsigned long long tol)
{
	int ret;
	/*
	 * First: Acquire the global semaphore and tasklist_lock
	 */
	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	ret = dump_hash_self(buf,out,tol);

	/*Last : Release the global semaphore and tasklist_lock*/
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);
	return ret;
}

/*
 * Wrapper of purge_hash_self w/ locking
 */
int purge_hash_other(ktau_prof *buf, 
		unsigned long long tol)
{
	int ret;
	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	ret = purge_hash_self(buf,tol);

	/*Last : Release the global semaphore and tasklist_lock*/
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);
	return ret;
}
#endif 
/*-------------------------------------------------------------------*/
/*
 * This is wrapped by dump_hash_many in case we are dumping all.
 */
#define for_each_process(p) for_each_task(p)

static int _dump_hash_all(ktau_output **arr,
		unsigned long long tol)
{
	struct task_struct *cur_tsk;
	int count = 0;
	int cur_count= 0;

	down_write(&ktau_global_sem);
	
	for_each_process(cur_tsk) count++;

	*arr = kmalloc(sizeof(ktau_output) * count, GFP_ATOMIC);	

	read_lock(&tasklist_lock);
	for_each_process(cur_tsk){
		(*arr+cur_count)->pid = cur_tsk->pid;
		if(dump_hash_self(cur_tsk->ktau, (*arr+cur_count), tol))
			break;
		cur_count++;
	}
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);

	return cur_count;
}

/*
 * This is wrapped by purge_hash_all in case we are purging all. 
 */
static int _purge_hash_all(unsigned long long tol)
{
	int cur_count = 0;
#if 0
	struct task_struct *cur_tsk;

	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);
	for_each_process(cur_tsk){
		if(purge_hash_self(cur_tsk->ktau, tol))
			break;
		cur_count++;
	}

	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);

#endif
	return cur_count;
}
/*-------------------------------------------------------------------*/
/*
 * dump_hash_many: Perform dump on other and all.
 */
int dump_hash_many(pid_t *pid_lst,
		unsigned int size, 
		ktau_output **arr, 
		unsigned long long tol)
{
	int ret_count = 0;
	int cur_count = 0;
	struct task_struct *cur_tsk;

	/* Check if passing ALL */
	if(!size){
		if(arr) return _dump_hash_all(arr,tol);
		else    return(-1);		/* Error arr is NULL */
	}else
		*arr = kmalloc(sizeof(ktau_output)*size,GFP_KERNEL);
	
	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	/* 
	 * Searching through the task list for the specified process.
	 * This should be optimized!!!!! Required the pid_list to be 
	 * sorted.
	 */

	for_each_process(cur_tsk){
		do{
			if(cur_tsk->pid == pid_lst[cur_count]){
				/* The specified PID is found */
				
				/* Read the pid from the process task_struct */	
				(*arr+ret_count)->pid = cur_tsk->pid;
			
				/* Do the same thing as dump_hash_self */
				if(dump_hash_self(cur_tsk->ktau, (*arr+ret_count), tol))
					break;

				/* Move to the next pid in the list */
				cur_count++;
				ret_count++;
			
			}else if(cur_tsk->pid > pid_lst[cur_count]){
				/* The specified PID is not found */
				/* Move to the next pid in the list */
				err("dump_hash_many : WARNING! PID %d not found.\n",
						pid_lst[cur_count]);	
				cur_count++;
			}
			/* Must check for array out-of-bound 
			 * before continue */
			if(cur_count >=  size || ret_count == size) 
				goto done;

		}while(cur_tsk->pid >= pid_lst[cur_count]);

	}
done:	
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);
	return ret_count;
}
/*
 * purge_hash_many: Perform purge on other and all.
 */
int purge_hash_many(pid_t *pid_lst,
		unsigned int size, 
		unsigned long long tol)
{
	int ret_count = 0;
#if 0
	int cur_count = 0;
	struct task_struct *cur_tsk;

	/* Check if passing ALL */
	if(!size) 
		return _purge_hash_all(tol);


	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	/* 
	 * Searching through the task list for the specified process.
	 * This should be optimized!!!!! Required the pid_list to be 
	 * sorted.
	 */
	for_each_process(cur_tsk){
		if(cur_tsk->pid == pid_lst[cur_count]){
			/* The specified PID is found */
			if(purge_hash_self(cur_tsk->ktau,tol))
				break;
			cur_count++;
			ret_count++;
		}else if(cur_tsk->pid > pid_lst[cur_count]){
			/* The specified PID is not found */
			cur_count++;
		}

		/* Check if we have worked  upto the size specified*/
		if(ret_count == size)
			break;
	}

	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);

#endif
	return ret_count;
}
/*--------------------------- CONTROL API ---------------------------------*/
/*
 * This function handles setting 'merge'-uspace-pointer 
 */
int set_hash_userstate(ktau_prof *buf, 
		ktau_state** ppstate) 
{
	unsigned long kflags = 0;
	int ret_value = 0;

	ktau_spin_lock(&buf->hash_lock,kflags);

	//info("ktau_hash.c:: set_hash_userstate: ppstate: %u\n",ppstate);

	if(ppstate != NULL)
	{
		//sanity check
		if(!access_ok(VERIFY_WRITE, ppstate, sizeof(ppstate))) 
		{
			//check failed
			err("ktau_hash.c:: set_hash_userstate: acess_ok failed.\n");
			ret_value = -1;
			goto unlock_out;
		}
	}
	
	//We allow setting ppstate to NULL (unset)
	buf->ppstate = ppstate;

	//info("ktau_hash.c:: set_hash_userstate: assigned: ktau->ppstate: %u\n",
	//					buf->ppstate);

	ret_value = 0;
	
unlock_out:

	ktau_spin_unlock(&buf->hash_lock,kflags);

	return ret_value;
}


/*
 * This function updates the uspace-state used 
 * for merging kernel & user profiles.
 *
 * Note: Does not perform ANY LOCKING. Hence any 
 * required locking must be performed externally.
 */
int update_userstate(unsigned long long incl, ktau_state** ppstate) {
	ktau_state kstate;
	ktau_state* kpstate = NULL;

	//info("update_userstate: ppstate: %u\n", ppstate);

	if(copy_from_user(&kpstate, ppstate, sizeof(kpstate))) {
		err("update_userstate: 1st copy_from_user failed.\n");
		return -1;
	}

	//info("update_userstate: kpstate: %u\n", kpstate);
	
	if(copy_from_user(&kstate, kpstate, sizeof(kstate))) {
		err("update_userstate: 2nd copy_from_user failed.\n");
		return -1;
	}

	info("update_userstate: ktime: %10llu knumcalls: %u\n", 
			kstate.ktime, kstate.knumcalls);

	kstate.ktime += incl;
	kstate.knumcalls += 1;

	if(copy_to_user(kpstate, &kstate, sizeof(kstate))) {
		err("update_userstate: copy_to_user failed.\n");
		return -1;
	}

	return 0;

}
