//kernel/ktau/ktau_proc_interface.c
//

/* ktau KERNEL headers */
#include <linux/ktau/ktau_proc_data.h>

/* ktau user-spc-interface headers */
#include <linux/ktau/ktau_proc_interface.h>

/* user-sp headers */
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

/* User-Space Application Interface Functions */


/* Read & Purge PIDs 				    *
 * If nopids = 1, then used for single pid          *
 * If nopids > 1, then used for multiple pids       *
 * if nopids = 0 & pid=null, then used for ALL Pids *
 */

#define MAX_KTAU_PROC_PATH 1024

/* read_size: Used to query /proc/ktau for SIZE of profile(s). *
 ***************************************************************
 * Arguments:
 * pid:         pointer to a single pid , or address of array of pids
 *              Can be NULL (if nopids is ZERO) --> size of FULL sys. prof
 *
 * nopids:      no of pids being pointed to above
 *              Can be ZERO --> request size of entire system prof
 *
 * tol:         tolerance paramter. to ignore, set to ZERO
 *
 * flags:       <TO BE DONE> . ignore, set to NULL.
 *
 * compratio:	This is a fraction between 0 & 1 (can be zero or 1 also)
 * 		-1 --> means USE DEFAULT VALUE
 * 		Compensation-Ratio Used to scale-up the size returned 
 * 		from /proc, so that when a call is made to read_data, 
 * 		the size is larger to accomodate any changes to profile. 
 * 		NOTE: To Ignore SET to -1. 
 *
 * Returns:
 * On Success:  total size of profile(s) of pid(s)
 * On Error:    -1; Other negative error codes reserved.
 *
 * Constraints:
 * None.
 */
long read_size(unsigned int *pid, unsigned int nopids, unsigned long tol, unsigned int* flags, float compratio)
{
	int ret_value = 0;
	int fd = -1;
	char path[MAX_KTAU_PROC_PATH+1];
	ktau_package_t ktau_pkg;
	ktau_size_t size;
	unsigned long lc_size = 0;

	/* initialize ktau_size_t with user-args */
	size.tol.tol_in = tol;
	size.flags = 0;
	if(flags)
		size.flags = *flags;

	if(nopids == 1) /* single pid get size : use the read method */
	{
		snprintf(path, MAX_KTAU_PROC_PATH, "/proc/ktau/%u", *pid);

		fd = open(path,O_RDWR);	
		if(fd < 0)
		{
			perror("open:");
			return fd;
		}

		ktau_pkg.op.size = size;
		ktau_pkg.type = CMD_SIZE;
		
		ret_value = 0;
		
		do {
			ret_value = write(fd, &ktau_pkg, sizeof(ktau_package_t));
		} while((ret_value < 0) && (errno == EINTR));

		lc_size = ktau_pkg.op.size.size;
		if(flags)
			*flags = ktau_pkg.op.size.flags;

	}
	else if (nopids == 0) /* ALL Pids: use IOCTL method */
	{
		snprintf(path, MAX_KTAU_PROC_PATH, "/proc/ktau/all");

		fd = open(path,O_RDWR);	
		if(fd < 0)
		{
			perror("open:");
			return fd;
		}
		
		ktau_pkg.op.size = size;
		ktau_pkg.type = CMD_ALL_SIZE;
		
		ret_value = ioctl(fd, CMD_ALL_SIZE, &ktau_pkg);
		lc_size = ktau_pkg.op.size.size;
		if(flags)
			*flags = ktau_pkg.op.size.flags;

	}
	else /* Many Pids: We dont handle it now */
	{
		perror("read_size: Dont Handle Multiple PIDs now. Only Single & ALL.\n");
		ret_value = -1;
	}

	close(fd);

	if(ret_value < 0)
	{
		perror("read:");
		return ret_value;
	}

	ret_value = 0;

	if(compratio == -1)
	{	
		compratio = 0.5;
	}

	lc_size = lc_size + (lc_size * compratio);

	return lc_size;
}


/* read_data: Used to query /proc/ktau for DATA of profile(s). *
 ***************************************************************
 * Arguments:
 * pid:         pointer to a single pid , or address of array of pids
 *              Can be NULL --> request data of entire system prof
 *
 * nopids:      no of pids being pointed to above
 *              Can be ZERO --> request data of entire system prof
 *
 * buffer:      pointer to allocated memory
 *
 * size:        size of buffer (allocate memory above)
 *
 * tol:         tolerance paramter. to ignore, set to ZERO
 *
 * flags:       <TO BE DONE> . ignore, set to NULL.
 *
 * Returns:
 * On Success:  total size of profile(s) data read into buffer.
 * On Error:    -1; Other negative error codes reserved.
 *
 * Constraints:
 * Must have called read_size before to ascertain size of buffer.
 *
 * NOTE: Even if SIZE allocated is that returned by read_size, 
 * read_data can return error for lack of size -> as size of prof
 * can change after read_size. This is unlikely.
 */
long read_data(unsigned int *pid, unsigned int nopids, char* buffer, unsigned long size, unsigned long tol, unsigned int* flags)
{
	int ret_value = 0;
	int fd = -1;
	char path[MAX_KTAU_PROC_PATH+1];
	ktau_package_t ktau_pkg;
	ktau_read_t kbuffer;
	char *lc_buf = NULL;
	char* temp_buf = NULL;
	unsigned int in_size = sizeof(ktau_package_t) + size ;

	//initialize buffer with user-args
	kbuffer.tol.tol_in = tol;
	kbuffer.data = buffer;
	kbuffer.size = size;
	kbuffer.flags = 0;
	if(flags)
		kbuffer.flags = *flags;

	lc_buf = (char*)malloc(in_size);
	if(!lc_buf)
	{
		perror("malloc: ");
		return -1;
	}

	if(nopids == 1) /* single pid get read : use the read method */
	{
		snprintf(path, MAX_KTAU_PROC_PATH, "/proc/ktau/%u", *pid);

		fd = open(path,O_RDWR);	
		if(fd < 0)
		{
			perror("open:");
			free(lc_buf);
			return fd;
		}

		ktau_pkg.op.read= kbuffer; /* so that flags/tolerance passed from app is recorded */
		ktau_pkg.type = CMD_READ;
		
		memcpy(lc_buf, &ktau_pkg, sizeof(ktau_pkg));

		ret_value = 0;
		do {
			ret_value = write(fd, lc_buf, sizeof(ktau_package_t) + kbuffer.size);
		} while((ret_value < 0) && (errno == EINTR));

	}
	else if(nopids == 0) /* ALL Pids : use IOCTL */
	{
		snprintf(path, MAX_KTAU_PROC_PATH, "/proc/ktau/all");

		fd = open(path,O_RDWR);	
		if(fd < 0)
		{
			perror("open:");
			free(lc_buf);
			return fd;
		}

		ktau_pkg.op.read= kbuffer; /* So that flags/tolerance passed from app is recorded */
		ktau_pkg.type = CMD_ALL_READ;

		memcpy(lc_buf, &ktau_pkg, sizeof(ktau_pkg));
		
		ret_value = ioctl(fd, CMD_ALL_READ, lc_buf);

	}
	else /* Mamny pids : We dont support */
	{
		perror("read_data: Many PIDs not supported. Only Single PID & ALL Pids.");
		ret_value = -1;
	}

	close(fd);

	if(ret_value < 0)
	{
		perror("read/ioctl ret val < 0:");
		if(lc_buf)
		{
			free(lc_buf);
		}
		return ret_value;
	}

	temp_buf = kbuffer.data;
	kbuffer = (*(ktau_package_t*)lc_buf).op.read;
	kbuffer.data = temp_buf;

	memcpy(kbuffer.data, (char*)(lc_buf + sizeof(ktau_package_t)), ret_value - sizeof(ktau_package_t));

	if(flags)
		*flags = kbuffer.flags;

	ret_value = 0;
	
	free(lc_buf);

	return kbuffer.size;
}


/* purge_data: Used to reset state of profiles /proc/ktau for pid(s). *
 **********************************************************************
 * Arguments:
 * pid:         pointer to a single pid , or address of array of pids
 *              Can be NULL --> request reset of entire system prof
 *
 * nopids:      no of pids being pointed to above
 *              Can be ZERO --> request reset of entire system prof
 *
 * tol:         tolerance paramter. to ignore, set to ZERO
 *
 * flags:       <TO BE DONE> . ignore, set to NULL.
 *
 * Returns:
 * On Success:  0
 * On Error:    -1; Other negative error codes reserved.
 *
 * Constraints:
 * None.
 */
long purge_data(unsigned int *pid, unsigned int nopids, unsigned long tol, unsigned int* flags)
{
	int ret_value = 0;
	int fd = -1;
	char path[MAX_KTAU_PROC_PATH+1];
	ktau_package_t ktau_pkg;
	ktau_write_t kbuffer;
	char* lc_buf = NULL;
	char* temp_buf = NULL;

	//initialize buffer with user-args
	kbuffer.tol.tol_in = tol;
	kbuffer.flags = 0;
	if(flags)
		kbuffer.flags = *flags;

	lc_buf = (char*)malloc(sizeof(ktau_package_t));
	if(!lc_buf)
	{
		perror("malloc:");
		return -1;
	}

	if(nopids == 1) /* single pid write : use the write method */
	{
		snprintf(path, MAX_KTAU_PROC_PATH, "/proc/ktau/%u", *pid);

		fd = open(path,O_RDWR);	
		if(fd < 0)
		{
			perror("open:");
			free(lc_buf);
			return fd;
		}

		ktau_pkg.op.write = kbuffer; /* so that flags/tolerance passed from app is recorded */
		ktau_pkg.type = CMD_PURGE;
		
		memcpy(lc_buf, &ktau_pkg, sizeof(ktau_pkg));

		ret_value = 0;
		do {
			ret_value = write(fd, lc_buf, sizeof(ktau_package_t));
		} while((ret_value < 0) && (errno == EINTR));

		ktau_pkg = *((ktau_package_t*)lc_buf);
	}
	else if(nopids == 0) /* ALL Pids : use IOCTL */
	{
		snprintf(path, MAX_KTAU_PROC_PATH, "/proc/ktau/all");

		fd = open(path,O_RDWR);	
		if(fd < 0)
		{
			perror("open:");
			free(lc_buf);
			return fd;
		}

		ktau_pkg.op.write = kbuffer; /* So that flags/ tolerance passed from app is recorded */
		ktau_pkg.type = CMD_ALL_PURGE;
		
		memcpy(lc_buf, &ktau_pkg, sizeof(ktau_pkg));

		ret_value = ioctl(fd, CMD_ALL_PURGE, lc_buf);

		ktau_pkg = *((ktau_package_t*)lc_buf);
	}
	else /*Many pids : not supported */
	{
		perror("pruge_data: Many PIDs not supported. Only Single & ALL Pids.");
		ret_value = -1;
	}

	close(fd);

	if(ret_value < 0)
	{
		perror("write / ioctl:");
		if(lc_buf)
		{
			free(lc_buf);
		}
		return ret_value;
	}

	kbuffer = ktau_pkg.op.write;

	if(flags)
		*flags = kbuffer.flags;

	ret_value = 0;
	
	free(lc_buf);

	return ret_value;
}


/* Once Read (into Buffer), that Data needs to be *
 * expanded into Profile Data.                    *
 * Returns: no profiles expanded. -1 on error.
 */
long unpack_bindata(char* buffer, unsigned long size, ktau_output** output) 
{
        int alloc_size = 0;
        int i = 0;
        char *bufptr = 0x0;
	int noprofs = -1;
	
	bufptr = buffer;

	memcpy(&noprofs, buffer, sizeof(unsigned int));	//<for no-of-profiles>
	bufptr+=sizeof(unsigned int);

	*output = (ktau_output*)malloc(noprofs * sizeof(ktau_output));
        if(!*output) {
                perror("malloc failed");
                return -1;
        }

        for(i = 0; i< noprofs; i++)
        {
		memcpy(&((*output)[i].pid), bufptr, sizeof(unsigned int)); //<for pid>
		bufptr+=sizeof(unsigned int);
		memcpy(&((*output)[i].size), bufptr, sizeof(unsigned int)); //<for size>
		bufptr+=sizeof(unsigned int);

		(*output)[i].ent_lst = (o_ent*)malloc(sizeof(o_ent)*(*output)[i].size);
        	if(!((*output)[i].ent_lst)) {
                	perror("malloc failed");
                	return -1;
        	}
		memcpy((*output)[i].ent_lst, bufptr, sizeof(o_ent) * (*output)[i].size); //<for o_ent>
		bufptr+= (sizeof(o_ent) * (*output)[i].size);
        }

        return noprofs;
}


/* Routine prints out an unpacked ktau_output       *
 * to the provided file-stream (can be stdout etc). *
 * Expects a const ptr to a single ktau_output */

/* Helper */
static void print_o_ent(FILE* fp, o_ent* pent);

void print_many_profiles(FILE* fp, const ktau_output* profiles, unsigned int no_profiles)
{
	int i = 0;

	fprintf(fp, "No Profiles: %u\n", no_profiles);
	for(i=0; i< no_profiles; i++)
	{
		print_ktau_output(fp, &(profiles[i]));
	}
}

void print_ktau_output(FILE* fp, const ktau_output* profile)
{
	o_ent* pent = NULL;
	int i = 0;

	if((!profile) || (!fp))
	{
		perror("Input Null ptr. \n");
		return;
	}

	fprintf(fp, "PID: %u\t No Entries: %u\n", profile->pid, profile->size);

	for(i=0; i<profile->size; i++)
	{
		pent = &(profile->ent_lst[i]);
		print_o_ent(fp, pent);
	}

}

/* helper to print o_ents */
static void print_o_ent(FILE* fp, o_ent* pent)
{
	h_ent* ph_ent = NULL;
	int i = 0;
	ktau_timer* timer;
	int index = 0;

	if(!fp || !pent)
	{
		perror("Null input. \n");
		return;
	}

	index = pent->index;

	ph_ent = &(pent->entry);

	if(!ph_ent)
	{
		fprintf(fp, "NULL\n");
		return;
	}

	timer = &(ph_ent->data.timer);

        //if(timer->count){
                fprintf(fp, "Entry %4d: addr %x, count %4u, total %4llu\n",
                                        index,
                                        timer->addr,
                                        timer->count,
                                        timer->total);
	//}

	return;
}


